/* $Header: /belch_a/users/rearl/tinymuck/src/RCS/look.c,v 1.17 90/10/06 19:04:30 rearl Exp $ */

/*
 * $Log:	look.c,v $
 * Revision 1.17  90/10/06  19:04:30  rearl
 * Fixed a typo in @# handling.
 * 
 * Revision 1.16  90/10/06  16:32:49  rearl
 * Fixes for 2.2 distribution.
 * 
 * Revision 1.15  90/09/28  12:24:13  rearl
 * Fixed small bug with disappearing descriptions...
 * 
 * Revision 1.14  90/09/18  08:00:02  rearl
 * Took out FILTER.  Fixed @find.
 * 
 * Revision 1.13  90/09/16  04:42:25  rearl
 * Preparation code added for disk-based MUCK.
 * 
 * Revision 1.12  90/09/15  22:25:59  rearl
 * Added argument passing to MUF from @#.  COMPRESS fix for this as well.
 * 
 * Revision 1.11  90/09/13  06:27:38  rearl
 * Argument passing with @# added.
 * 
 * Revision 1.10  90/08/27  03:29:34  rearl
 * Added environment support.
 * 
 * Revision 1.9  90/08/11  04:05:21  rearl
 * *** empty log message ***
 * 
 * Revision 1.8  90/08/06  02:35:27  rearl
 * Oopsie, changed controls() back to can_link().
 * 
 * Revision 1.7  90/08/05  03:19:33  rearl
 * Redid matching routines.
 * 
 * Revision 1.6  90/08/02  18:53:09  rearl
 * Changed can_link() to controls().  Everyone controls an unlinked exit.
 * 
 * Revision 1.5  90/07/30  00:09:55  rearl
 * Added @owned command, lists everything you own, or for wizards,
 * everything owned by the specified player (including exits)
 * 
 * Revision 1.4  90/07/29  21:19:17  rearl
 * Fixed examine-description brain damage.
 * 
 * Revision 1.3  90/07/29  17:39:30  rearl
 * Fixed various things such as ROOM programs, examine bug, examine
 * of programs now shows location, etc.
 * 
 * Revision 1.2  90/07/23  03:12:10  casie
 * Cleaned up various gcc warnings.
 * 
 * Revision 1.1  90/07/19  23:03:52  casie
 * Initial revision
 * 
 *
 */

#include "copyright.h"
#include "config.h"

/* commands which look at things */

#include <ctype.h>
#include "db.h"
#include "params.h"
#include "interface.h"
#include "match.h"
#include "externs.h"

#ifdef COMPRESS
extern const char *uncompress(const char *);
#endif /* COMPRESS */
#define EXEC_SIGNAL '@'    /* Symbol which tells us what we're looking at
			    * is an execution order and not a message.    */

/* prints owner of something */
static void
  print_owner(dbref player, dbref thing)
{
  char buf[BUFFER_LEN];
  
  switch(Typeof(thing))
    {
    case TYPE_PLAYER:
      sprintf(buf, "%s is a player.", NAME(thing));
      break;
    case TYPE_ROOM:
    case TYPE_THING:
    case TYPE_EXIT:
    case TYPE_PROGRAM:
      sprintf(buf, "Owner: %s", NAME(OWNER(thing)));
      break;
#ifdef RECYCLE
    case TYPE_GARBAGE:
      sprintf(buf, "%s is garbage.", NAME(thing));
      break;
#endif      
    }
  notify(player, buf);
}  

void exec_or_notify(dbref player, dbref thing, const char *message)
{
  const char *p;
#ifdef COMPRESS
  p = uncompress(message);
#else /* !COMPRESS */
  p = message;
#endif /* COMPRESS */
  if (*p == EXEC_SIGNAL) {
    int i;

    i = atoi(++p);
    for (; *p && !isspace(*p); p++)
      ;
    if (*p) p++;
    if ((Typeof(i) != TYPE_PROGRAM) || DBFETCH(player)->sp.player.run) {
      if (*p) notify (player, p);
      else notify(player, "You see nothing special.");
    } else {
      strcpy(match_args, p);
      (void) interp(player, i, thing);
      if (!(DBFETCH(player)->sp.player.run -> pc)) {
	free((void *) DBFETCH(player)->sp.player.run);
	DBSTORE(player, sp.player.run, 0);
      }
    }
  } else notify (player, p);
}

static void look_contents(dbref player, dbref loc, const char *contents_name)
{
  dbref thing;
  dbref can_see_loc;
  
  /* check to see if he can see the location */
  can_see_loc = (!Dark(loc) || controls(player, loc));
  
  /* check to see if there is anything there */
  DOLIST(thing, DBFETCH(loc)->contents) {
    if(can_see(player, thing, can_see_loc)) {
      /* something exists!  show him everything */
      notify(player, contents_name);
      DOLIST(thing, DBFETCH(loc)->contents) {
	if(can_see(player, thing, can_see_loc)) {
	  notify(player, unparse_object(player, thing));
	}
      }
      break;		/* we're done */
    }
  }
}

static void look_simple(dbref player, dbref thing)
{
  if(DBFETCH(thing)->description) {
    exec_or_notify(player, thing, DBFETCH(thing)->description);
  } else {
    notify(player, "You see nothing special.");
  }
}

void look_room(dbref player, dbref loc)
{
  /* tell him the name, and the number if he can link to it */
  notify(player, unparse_object(player, loc));
  
  /* tell him the description */
  if(DBFETCH(loc)->description) {
    exec_or_notify(player, loc, DBFETCH(loc)->description);
  }
  
  /* tell him the appropriate messages if he has the key */
  can_doit(player, loc, 0);
  
  /* tell him the contents */
  look_contents(player, loc, "Contents:");
}

void do_look_around(dbref player)
{
  dbref loc;
  
  if((loc = getloc(player)) == NOTHING) return;
  look_room(player, loc);
}

void do_look_at(dbref player, const char *name)
{
  dbref thing;
  struct match_data md;
  
  if(*name == '\0') {
    if((thing = getloc(player)) != NOTHING) {
      look_room(player, thing);
    }
  } else {
    /* look at a thing here */
    init_match(player, name, NOTYPE, &md);
    match_all_exits(&md);
    match_neighbor(&md);
    match_possession(&md);
    if(Wizard(player)) {
      match_absolute(&md);
      match_player(&md);
    }
    match_here(&md);
    match_me(&md);
    
    if((thing = noisy_match_result(&md)) != NOTHING) {
      switch(Typeof(thing)) {
      case TYPE_ROOM:
	if (getloc(player) != thing
	    && !can_link_to(player, TYPE_ROOM, thing)) {
	  notify(player, "Permission denied.");
	} else {
	  look_room(player, thing);
	}
	break;
      case TYPE_PLAYER:
	look_simple(player, thing);
	look_contents(player, thing, "Carrying:");
	break;
      default:
	look_simple(player, thing);
	break;
      }
    }
  }
}

#ifdef VERBOSE_EXAMINE
static const char *flag_description(dbref thing)
{
  static char buf[BUFFER_LEN];
  
  strcpy(buf, "Type: ");
  switch(Typeof(thing)) {
  case TYPE_ROOM:
    strcat(buf, "ROOM");
    break;
  case TYPE_EXIT:
    strcat(buf, "EXIT/ACTION");
    break;
  case TYPE_THING:
    strcat(buf, "THING");
    break;
  case TYPE_PLAYER:
    strcat(buf, "PLAYER");
    break;
  case TYPE_PROGRAM:
    strcat(buf, "PROGRAM");
    break;
#ifdef RECYCLE	
  case TYPE_GARBAGE:
    strcat(buf, "GARBAGE");
    break;
#endif	
  default:
    strcat(buf, "***UNKNOWN TYPE***");
    break;
  }
  
  if(FLAGS(thing) & ~TYPE_MASK) {
    /* print flags */
    strcat(buf, "  Flags:");
    if(FLAGS(thing) & WIZARD) strcat(buf, " WIZARD");
    if(FLAGS(thing) & STICKY)
      strcat(buf, (Typeof(thing) != TYPE_PROGRAM) ? " STICKY" : " SETUID");
    if(FLAGS(thing) & DARK)
      strcat(buf, (Typeof(thing) != TYPE_PROGRAM) ? " DARK" : " DEBUGGING");
    if(FLAGS(thing) & LINK_OK) strcat(buf, " LINK_OK");
#ifndef MUCKER_ALL
    if(FLAGS(thing) & MUCKER) strcat(buf, " MUCKER");
#endif /* MUCKER_ALL */
#ifdef RESTRICTED_BUILDING
    if(FLAGS(thing) & BUILDER) strcat(buf, " BUILDER");
#endif /* RESTRICTED_BUILDING */
#ifdef PLAYER_CHOWN
    if(FLAGS(thing) & CHOWN_OK) strcat(buf, " CHOWN_OK");
#endif /* PLAYER_CHOWN */
    if(FLAGS(thing) & JUMP_OK) strcat(buf, " JUMP_OK");
#ifdef HAVEN
    if (FLAGS(thing) & HAVEN) strcat(buf, " HAVEN");
#endif /* HAVEN */
#ifdef ABODE
    if (FLAGS(thing) & ABODE) strcat(buf, " ABODE");
#endif /* ABODE */
  }
  
  return buf;
}
#endif /* VERBOSE_EXAMINE */

void do_examine(dbref player, const char *name)
{
  dbref thing;
  char buf[BUFFER_LEN];
  dbref content;
  dbref exit;
  int  i;
  struct match_data md;
  
  if(*name == '\0') {
    if((thing = getloc(player)) == NOTHING) return;
  } else {
    /* look it up */
    init_match(player, name, NOTYPE, &md);
    match_all_exits(&md);
    match_neighbor(&md);
    match_possession(&md);
    match_absolute(&md);
    /* only Wizards can examine other players */
    if(Wizard(player)) match_player(&md);
    match_here(&md);
    match_me(&md);
    
    /* get result */
    if((thing = noisy_match_result(&md)) == NOTHING) return;
  }
  
  if(!can_link(player, thing)) {
    print_owner(player, thing);
    return;
  }

  switch(Typeof(thing)) {
  case TYPE_ROOM:
    sprintf(buf, "%s  Owner: %s  Parent: ", unparse_object(player, thing),
	    NAME(OWNER(thing)));
    strcat(buf, unparse_object(player, DBFETCH(thing)->location));
    break;
  case TYPE_THING:
    sprintf(buf, "%s  Owner: %s  Value: %d", unparse_object(player, thing),
	    NAME(OWNER(thing)), DBFETCH(thing)->sp.thing.value);
    break;
  case TYPE_PLAYER:
    sprintf(buf, "%s  Pennies: %d  ", unparse_object(player, thing),
	    DBFETCH(thing)->sp.player.pennies);
    break;
  case TYPE_EXIT:
  case TYPE_PROGRAM:
    sprintf(buf, "%s  Owner: %s", unparse_object(player, thing),
	    NAME(OWNER(thing)));
    break;
#ifdef RECYCLE
  case TYPE_GARBAGE:
    strcpy(buf, unparse_object(player, thing));
    break;
#endif	
  }
  notify(player, buf);
#ifdef VERBOSE_EXAMINE
  notify(player, flag_description(thing));
#endif /* VERBOSE_EXAMINE */
  if(DBFETCH(thing)->description) notify(player, DBFETCH(thing)->description);
  sprintf(buf, "Key: %s", unparse_boolexp(player, DBFETCH(thing)->key));
  notify(player, buf);
  
  if(DBFETCH(thing)->fail_message) {
    sprintf(buf, "Fail: %s", DBFETCH(thing)->fail_message);
    notify(player, buf);
  }
  if(DBFETCH(thing)->succ_message) {
    sprintf(buf, "Success: %s", DBFETCH(thing)->succ_message);
    notify(player, buf);
  }
  if (DBFETCH(thing)->drop_message)
    {
      sprintf(buf, "Drop: %s", DBFETCH(thing)->drop_message);
      notify(player, buf);
    }
  if(DBFETCH(thing)->ofail) {
    sprintf(buf, "Ofail: %s", DBFETCH(thing)->ofail);
    notify(player, buf);
  }
  if(DBFETCH(thing)->osuccess) {
    sprintf(buf, "Osuccess: %s", DBFETCH(thing)->osuccess);
    notify(player, buf);
  }
  if (DBFETCH(thing)->odrop)
    {
      sprintf(buf, "Odrop: %s", DBFETCH(thing)->odrop);
      notify(player, buf);
    }
  
  /* show him the properties */
  if (DBFETCH(thing)->properties)
    {
      struct plist *p;
      char   buf[BUFFER_LEN];
      
      notify(player, "Properties:");
      for (p = DBFETCH(thing)->properties; p; p = p -> next)
	{
	  strncpy(buf, p -> type, BUFFER_LEN - 1);
	  strncat(buf, ": ", BUFFER_LEN - 1);
	  if (p -> class)
	    strncat(buf, p -> class, BUFFER_LEN - 1);
#ifdef SHOWINTPROP
	  else
	    {
	      char xbuf[BUFFER_LEN];
	      
	      sprintf(xbuf, "%d", p -> value);
	      strncat(buf, xbuf, BUFFER_LEN - 1);
	    }
#endif /* SHOWINTPROP */
	  notify(player, buf);
	}
    }
  
  /* show him the contents */
  if(DBFETCH(thing)->contents != NOTHING) {
    if (Typeof(thing) == TYPE_PLAYER)
      notify(player, "Carrying:");
    else
      notify(player, "Contents:");
    DOLIST(content, DBFETCH(thing)->contents) {
      notify(player, unparse_object(player, content));
    }
  }
  
  switch(Typeof(thing)) {
  case TYPE_ROOM:
    /* tell him about exits */
    if(DBFETCH(thing)->sp.room.exits != NOTHING) {
      notify(player, "Exits:");
      DOLIST(exit, DBFETCH(thing)->sp.room.exits) {
	notify(player, unparse_object(player, exit));
      }
    } else {
      notify(player, "No exits.");
    }
    
    /* print dropto if present */
    if(DBFETCH(thing)->sp.room.dropto != NOTHING) {
      sprintf(buf, "Dropped objects go to: %s",
	      unparse_object(player, DBFETCH(thing)->sp.room.dropto));
      notify(player, buf);
    }
    break;
  case TYPE_THING:
    /* print home */
    sprintf(buf, "Home: %s",
	    unparse_object(player, DBFETCH(thing)->sp.thing.home)); /* home */
    notify(player, buf);
    /* print location if player can link to it */
    if(DBFETCH(thing)->location != NOTHING && (controls(player, DBFETCH(thing)->location)
	   || can_link_to(player, NOTYPE, DBFETCH(thing)->location))) {
      sprintf(buf, "Location: %s",
	      unparse_object(player, DBFETCH(thing)->location));
      notify(player, buf);
    }
    /* print thing's actions, if any */
    if(DBFETCH(thing)->sp.thing.actions != NOTHING) {
      notify(player, "Actions/exits:");
      DOLIST(exit, DBFETCH(thing)->sp.thing.actions) {
	notify(player, unparse_object(player, exit));
      }
    } else {
      notify(player, "No actions attached.");
    }
    break;
  case TYPE_PLAYER:
    
    /* print home */
    sprintf(buf, "Home: %s",
	    unparse_object(player, DBFETCH(thing)->sp.player.home)); /* home */
    notify(player, buf);
    
    /* print location if player can link to it */
    if(DBFETCH(thing)->location != NOTHING && (controls(player, DBFETCH(thing)->location)
	   || can_link_to(player, NOTYPE, DBFETCH(thing)->location))) {
      sprintf(buf, "Location: %s",
	      unparse_object(player, DBFETCH(thing)->location));
      notify(player, buf);
    }
    
    /* print player's actions, if any */
    if(DBFETCH(thing)->sp.player.actions != NOTHING) {
      notify(player, "Actions/exits:");
      DOLIST(exit, DBFETCH(thing)->sp.player.actions) {
	notify(player, unparse_object(player, exit));
      }
    } else {
      notify(player, "No actions attached.");
    }
    break;
  case TYPE_EXIT:
    if (DBFETCH(thing)->location != NOTHING) {
      sprintf(buf, "Source: %s", unparse_object(player, DBFETCH(thing)->location));
      notify(player, buf);
    }
    /* print destinations */
    if (DBFETCH(thing)->sp.exit.ndest == 0) break;
    for (i = 0; i < DBFETCH(thing)->sp.exit.ndest; i++) {
      switch( (DBFETCH(thing)->sp.exit.dest)[i]) {
      case NOTHING:
	break;
      case HOME:
	notify(player, "Destination: *HOME*");
	break;
      default:
	sprintf(buf, "Destination: %s",
		unparse_object(player, (DBFETCH(thing)->sp.exit.dest)[i]));
	notify(player, buf);
	break;
      }
    }
    break;
  case TYPE_PROGRAM:
    if (DBFETCH(thing)->sp.program.siz)
      sprintf(buf, "Program compiled size: %d", DBFETCH(thing)->sp.program.siz);
    else
      sprintf(buf, "Program not compiled.");
    notify(player, buf);
    
    /* print location if player can link to it */
    if(DBFETCH(thing)->location != NOTHING && (controls(player, DBFETCH(thing)->location)
	   || can_link_to(player, NOTYPE, DBFETCH(thing)->location))) {
      sprintf(buf, "Location: %s", unparse_object(player, DBFETCH(thing)->location));
      notify(player, buf);
    }
    break;
  default:
    /* do nothing */
    break;
  }
}

void do_score(dbref player) 
{
  char buf[BUFFER_LEN];
  
  sprintf(buf, "You have %d %s.", DBFETCH(player)->sp.player.pennies,
	  DBFETCH(player)->sp.player.pennies == 1 ? "penny" : "pennies");
  notify(player, buf);
}

void do_inventory(dbref player)
{
  dbref thing;
  
  if((thing = DBFETCH(player)->contents) == NOTHING) {
    notify(player, "You aren't carrying anything.");
  } else {
    notify(player, "You are carrying:");
    DOLIST(thing, thing) {
      notify(player, unparse_object(player, thing));
    }
  }
  
  do_score(player);
}

void do_find(dbref player, const char *name)
{
  dbref i;
  
  if(!payfor(player, LOOKUP_COST)) {
    notify(player, "You don't have enough pennies.");
  } else {
    for(i = 0; i < db_top; i++) {
      if(Typeof(i) != TYPE_EXIT && controls(player, i)
	 && (!*name || string_match(NAME(i), name))) {
	notify(player, unparse_object(player, i));
      }
    }
    notify(player, "***End of List***");
  }
}

void do_owned(dbref player, const char *name)
{
  dbref victim, i;
  
  if(!payfor(player, LOOKUP_COST)) {
    notify(player, "You don't have enough pennies.");
    return;
  }
  if (Wizard(player) && *name) {
    if ((victim = lookup_player(name)) == NOTHING) {
      notify(player, "I couldn't find that player.");
      return;
    }
  }
  else victim = player;
  for(i = 0; i < db_top; i++) {
    if (OWNER(i) == victim)
      notify(player, unparse_object(player, i));
  }
  notify(player, "***End of List***");
}

void do_trace(dbref player, const char *name, int depth)
{
  dbref thing;
  int i;
  struct match_data md;

  init_match(player, name, NOTYPE, &md);
  match_absolute(&md);
  match_here(&md);
  match_me(&md);
  match_neighbor(&md);
  match_possession(&md);
  if ((thing = noisy_match_result(&md)) == NOTHING
      || thing == AMBIGUOUS) return;

  for (i = 0; (!depth || i < depth) && thing != NOTHING; i++) {
    if (controls(player, thing) || can_link_to(player, NOTYPE, thing))
      notify(player, unparse_object(player, thing));
    else
      notify(player, "**Missing**");
    thing = DBFETCH(thing)->location;
  }
  notify(player, "***End of List***");
}
