/* $Header: /belch_a/users/rearl/tinymuck/src/RCS/predicates.c,v 1.8 90/09/18 08:01:39 rearl Exp $ */

/*
 * $Log:	predicates.c,v $
 * Revision 1.8  90/09/18  08:01:39  rearl
 * Fixed a few broken things.
 * 
 * Revision 1.7  90/09/16  04:42:42  rearl
 * Preparation code added for disk-based MUCK.
 * 
 * Revision 1.6  90/09/10  02:22:30  rearl
 * Fixed some calls to pronoun_substitute.
 * 
 * Revision 1.5  90/08/27  03:32:49  rearl
 * Major changes on several predicates, usage...
 * 
 * Revision 1.4  90/08/15  03:07:34  rearl
 * Macroified some things.  Took out #ifdef GENDER.
 * 
 * Revision 1.3  90/08/06  02:46:30  rearl
 * Put can_link() back in.  Added restricted() for flags.
 * With #define GOD_PRIV, sub-wizards can no longer set other
 * players to Wizard.
 * 
 * Revision 1.2  90/08/02  18:54:04  rearl
 * Fixed controls() to return TRUE if exit is unlinked.  Everyone
 * controls an unlinked exit.  Got rid of can_link().
 * 
 * Revision 1.1  90/07/19  23:03:58  casie
 * Initial revision
 * 
 *
 */

#include "copyright.h"
#include "config.h"

/* Predicates for testing various conditions */

#include <ctype.h>

#include "db.h"
#include "interface.h"
#include "params.h"
#include "externs.h"

int can_link_to(dbref who, object_flag_type what_type, dbref where)
{
  if (where == HOME) return 1;
  if (where < 0 || where > db_top) return 0;
  switch (what_type) {
  case TYPE_EXIT:
    return (controls(who, where) || (FLAGS(where) & LINK_OK));
    /*NOTREACHED*/
    break;
  case TYPE_PLAYER:
    return (Typeof(where) == TYPE_ROOM && (controls(who, where)
					   || Linkable(where)));
    /*NOTREACHED*/
    break;
  case TYPE_ROOM:
  case TYPE_THING:
    return (Typeof(where) == TYPE_ROOM && (controls(who, where)
					   || Linkable(where)));
    /*NOTREACHED*/
    break;
  case NOTYPE:
#ifdef ABODE
    return (controls(who, where) || (FLAGS(where) & (ABODE|LINK_OK)));
#else /* ABODE */
    return (controls(who, where) || (FLAGS(where) & LINK_OK));
#endif /* ABODE */
    /*NOTREACHED*/
    break;
  }
  return 0;
}

int can_link(dbref who, dbref what)
{
  return (controls(who, what) || ((Typeof(what) == TYPE_EXIT)
				  && DBFETCH(what)->sp.exit.ndest == 0));
}

int could_doit(dbref player, dbref thing)
{
  if(Typeof(thing) == TYPE_EXIT && DBFETCH(thing)->sp.exit.ndest == 0) {
    return 0;
  }
  return(eval_boolexp (player, DBFETCH(thing)->key, thing));
}

int can_doit(dbref player, dbref thing, const char *default_fail_msg)
{
  dbref loc;
  char buf[BUFFER_LEN];
  
  if((loc = getloc(player)) == NOTHING) return 0;
  
  if(!could_doit(player, thing)) {
    /* can't do it */
    if(DBFETCH(thing)->fail_message) {
      exec_or_notify(player, thing, DBFETCH(thing)->fail_message);
    } else if(default_fail_msg) {
      notify(player, default_fail_msg);
    }
    
    if(DBFETCH(thing)->ofail && !Dark(player)) {
      sprintf(buf, "%s %s", NAME(player),
	      pronoun_substitute(player, DBFETCH(thing)->ofail));
      notify_except(DBFETCH(loc)->contents, player, buf);
    }
    
    return 0;
  } else {
    /* can do it */
    if(DBFETCH(thing)->succ_message) {
      exec_or_notify(player, thing, DBFETCH(thing)->succ_message);
    }
    
    if(DBFETCH(thing)->osuccess && !Dark(player)) {
      sprintf(buf, "%s %s", NAME(player),
	      pronoun_substitute(player, DBFETCH(thing)->osuccess));
      notify_except(DBFETCH(loc)->contents, player, buf);
    }
    
    return 1;
  }
}

int can_see(dbref player, dbref thing, int can_see_loc)
{
  if (player == thing || Typeof(thing) == TYPE_EXIT
      || Typeof(thing) == TYPE_ROOM) return 0;

  if (can_see_loc) {
    switch (Typeof(thing)) {
    case TYPE_PROGRAM:
      return((FLAGS(thing) & LINK_OK) || controls(player, thing));
    default:
      return (!Dark(thing) || controls(player, thing));
    }
  } else {
    /* can't see loc */
    return(controls(player, thing));
  }
}

int controls(dbref who, dbref what)
{
  /* Wizard controls everything */
  /* owners control their stuff */
  if (what < 0 || what >= db_top) return 0;
#ifdef RECYCLE
  /* No one controls garbage */
  if (Typeof(what) == TYPE_GARBAGE) return 0;
#endif    
  if (Wizard(who)) return 1;
  return (who == OWNER(what));
}

int restricted(dbref player, dbref thing, object_flag_type flag)
{
  switch (flag) {
  case DARK:
    return (!Wizard(player) && (Typeof(thing) != TYPE_ROOM)
	    && (Typeof(thing) != TYPE_PROGRAM)); /* Debugging */
    /*NOTREACHED*/
    break;
#ifndef MUCKER_ALL
  case MUCKER:
#endif /* MUCKER_ALL */
#ifdef RESTRICTED_BUILDING
  case BUILDER:
#endif /* RESTRICTED_BUILDING */
    return (!Wizard(player));
    /*NOTREACHED*/
    break;
  case WIZARD:
    if (Wizard(player)) {
#ifdef GOD_PRIV
      return ((Typeof(thing) == TYPE_PLAYER) && !God(player));
#else /* !GOD_PRIV */
      return 0;
#endif /* GOD_PRIV */
    } else return 1;
    /*NOTREACHED*/
    break;
  default:
    return 0;
    /*NOTREACHED*/
    break;
  }
}

int payfor(dbref who, int cost)
{
  if(Wizard(who)) {
    return 1;
  } else if(DBFETCH(who)->sp.player.pennies >= cost) {
    DBFETCH(who)->sp.player.pennies -= cost;
    DBDIRTY(who);
    return 1;
  } else {
    return 0;
  }
}
int word_start (const char *str, const char let)
{
  int chk;
  
  for (chk = 1; *str; str++) {
    if (chk && *str == let) return 1;
    chk = *str == ' ';
  }
  return 0;
}

int ok_name(const char *name)
{
  return (name
	  && *name
	  && *name != LOOKUP_TOKEN
	  && *name != NUMBER_TOKEN
	  && !index(name, ARG_DELIMITER)
	  && !index(name, AND_TOKEN)
	  && !index(name, OR_TOKEN)
	  && !word_start(name, NOT_TOKEN)
	  && string_compare(name, "me")
	  && string_compare(name, "home")
	  && string_compare(name, "here"));
}

int ok_player_name(const char *name)
{
  const char *scan;
  
  if(!ok_name(name) || strlen(name) > PLAYER_NAME_LIMIT) return 0;
  
  for(scan = name; *scan; scan++) {
    if(!(isprint(*scan) && !isspace(*scan))) { /* was isgraph(*scan) */
      return 0;
    }
  }
  
  /* lookup name to avoid conflicts */
  return (lookup_player(name) == NOTHING);
}

int ok_password(const char *password)
{
  const char *scan;
  
  if(*password == '\0') return 0;
  
  for(scan = password; *scan; scan++) {
    if(!(isprint(*scan) && !isspace(*scan))) {
      return 0;
    }
  }
  
  return 1;
}
