/* $Header: /belch_a/users/rearl/tinymuck/src/RCS/set.c,v 1.13 90/09/18 08:02:06 rearl Exp $ */

/*
 * $Log:	set.c,v $
 * Revision 1.13  90/09/18  08:02:06  rearl
 * Player hash table mods, took out FILTER.
 * 
 * Revision 1.12  90/09/16  04:42:56  rearl
 * Preparation code added for disk-based MUCK.
 * 
 * Revision 1.11  90/09/10  02:23:11  rearl
 * Took out get_string call.
 * 
 * Revision 1.10  90/08/27  03:33:18  rearl
 * Fixed bug involving NULL player password entries.
 * 
 * Revision 1.9  90/08/11  04:10:15  rearl
 * *** empty log message ***
 * 
 * Revision 1.8  90/08/06  03:51:35  rearl
 * Redid @chown thing = me, the easy way...
 * 
 * Revision 1.7  90/08/06  02:51:18  rearl
 * Added restricted() routine from predicates.c, any restricted
 * flags should be checked in there from now on.
 * 
 * Revision 1.6  90/08/05  03:20:01  rearl
 * Redid matching routines.
 * 
 * Revision 1.5  90/08/02  18:51:26  rearl
 * Enabled @chown <thing> = me in addition to @chown <thing>.
 * 
 * Revision 1.4  90/07/29  17:43:30  rearl
 * Allows players to capitalize/decapitalize their names, added support
 * for setting programs DEBUG (DARK).
 * 
 * 
 * Revision 1.3  90/07/21  16:31:50  casie
 * added log_misc
 * 
 * Revision 1.2  90/07/21  05:32:41  casie
 * Changed player chown to allow rooms and exits as well as things.
 * 
 * Revision 1.1  90/07/19  23:04:09  casie
 * Initial revision
 * 
 *
 */

#include "copyright.h"
#include "config.h"

/* commands which set parameters */
#include <stdio.h>
#include <ctype.h>
#include "strings.h"

#include "db.h"
#include "params.h"
#include "match.h"
#include "interface.h"
#include "externs.h"

#ifdef COMPRESS
#define alloc_compressed(x) alloc_string(compress(x))
#else /* COMPRESS */
#define alloc_compressed(x) alloc_string(x)
#endif /* COMPRESS */

static dbref match_controlled(dbref player, const char *name)
{
  dbref match;
  struct match_data md;
  
  init_match(player, name, NOTYPE, &md);
  match_everything(&md);
  
  match = noisy_match_result(&md);
  if(match != NOTHING && !controls(player, match)) {
    notify(player, "Permission denied.");
    return NOTHING;
  } else {
    return match;
  }
}

void do_name(dbref player, const char *name, char *newname)
{
  dbref thing;
  char *password;
  
  if((thing = match_controlled(player, name)) != NOTHING) {
    /* check for bad name */
    if(*newname == '\0') {
      notify(player, "Give it what new name?");
      return;
    }

    /* check for renaming a player */
    if(Typeof(thing) == TYPE_PLAYER) {
      /* split off password */
      for(password = newname;
	  *password && !isspace(*password);
	  password++);
      /* eat whitespace */
      if(*password) {
	*password++ = '\0'; /* terminate name */
	while(*password && isspace(*password)) password++;
      }
      
      /* check for null password */
      if(!*password) {
	notify(player,
	       "You must specify a password to change a player name.");
	notify(player, "E.g.: name player = newname password");
	return;
      } else if(strcmp(password, DoNull(DBFETCH(thing)->sp.player.password))) {
	notify(player, "Incorrect password.");
	return;
      } else if(string_compare(newname, NAME(thing))
		&& !ok_player_name(newname)) {
	notify(player, "You can't give a player that name.");
	return;
      }
      /* everything ok, notify */
      log_status("NAME CHANGE: %s(#%d) to %s\n",
		 NAME(thing), thing, newname);
      delete_player(thing);
      if (NAME(thing)) free((void *) NAME(thing));
      NAME(thing) = alloc_string(newname);
      add_player(thing);
      notify(player, "Name set.");
      return;
    } else {
      if(!ok_name(newname)) {
	notify(player, "That is not a reasonable name.");
	return;
      }
    }
    
    /* everything ok, change the name */
    if(NAME(thing)) {
      free((void *) NAME(thing));
    }
    NAME(thing) = alloc_string(newname);
    notify(player, "Name set.");
    DBDIRTY(thing);
  }
}

void do_describe(dbref player, const char *name, const char *description)
{
  dbref thing;
  
  if((thing = match_controlled(player, name)) != NOTHING) {
    if(DBFETCH(thing)->description) {
      free((void *) DBFETCH(thing)->description);
    }
    DBSTORE(thing, description, alloc_compressed(description));
    notify(player, "Description set.");
  }
}

void do_fail(dbref player, const char *name, const char *message)
{
  dbref thing;
  
  if((thing = match_controlled(player, name)) != NOTHING) {
    if(DBFETCH(thing)->fail_message) {
      free((void *) DBFETCH(thing)->fail_message);
    }
    DBSTORE(thing, fail_message, alloc_compressed(message));
    notify(player, "Message set.");
  }
}

void do_success(dbref player, const char *name, const char *message)
{
  dbref thing;
  
  if((thing = match_controlled(player, name)) != NOTHING) {
    if(DBFETCH(thing)->succ_message) {
      free((void *) DBFETCH(thing)->succ_message);
    }
    DBSTORE(thing, succ_message, alloc_compressed(message));
    notify(player, "Message set.");
  }
}

/* sets the drop message for player */
void
  do_drop_message(dbref player, const char *name, const char *message)
{
  dbref thing;
  if ((thing = match_controlled(player, name)) != NOTHING)
    {
      if (DBFETCH(thing)->drop_message)
	{
	  free((void *) DBFETCH(thing)->drop_message);
	}
      DBSTORE(thing, drop_message, alloc_compressed(message));
      notify(player, "Message set.");
    }
}

void do_osuccess(dbref player, const char *name, const char *message)
{
  dbref thing;
  
  if((thing = match_controlled(player, name)) != NOTHING) {
    if(DBFETCH(thing)->osuccess) {
      free((void *) DBFETCH(thing)->osuccess);
    }
    DBSTORE(thing, osuccess, alloc_compressed(message));
    notify(player, "Message set.");
  }
}

void do_ofail(dbref player, const char *name, const char *message)
{
  dbref thing;
  
  if((thing = match_controlled(player, name)) != NOTHING) {
    if(DBFETCH(thing)->ofail) {
      free((void *) DBFETCH(thing)->ofail);
    }
    DBSTORE(thing, ofail, alloc_compressed(message));
    notify(player, "Message set.");
  }
}

void
  do_odrop(dbref player,const char *name, const char *message)
{
  dbref thing;
  
  if ((thing = match_controlled(player, name)) != NOTHING)
    {
      if (DBFETCH(thing)->odrop)
	{
	  free((void *) DBFETCH(thing)->odrop);
	}
      DBSTORE(thing, odrop, alloc_compressed(message));
      notify(player, "Message set.");
    }
}

void do_lock(dbref player, const char *name, const char *keyname)
{
  dbref thing;
  struct boolexp *key;
  struct match_data md;
  
  init_match(player, name, NOTYPE, &md);
  match_everything(&md);
  
  switch(thing = match_result(&md)) {
  case NOTHING:
    notify(player, "I don't see what you want to lock!");
    return;
  case AMBIGUOUS:
    notify(player, "I don't know which one you want to lock!");
    return;
  default:
    if(!controls(player, thing)) {
      notify(player, "You can't lock that!");
      return;
    }
    break;
  }
  
  key = parse_boolexp(player, keyname);
  if(key == TRUE_BOOLEXP) {
    notify(player, "I don't understand that key.");
  } else {
    /* everything ok, do it */
    free_boolexp(DBFETCH(thing)->key);
    DBSTORE(thing, key, key);
    notify(player, "Locked.");
  }
}

void do_unlock(dbref player, const char *name)
{
  dbref thing;
  
  if((thing = match_controlled(player, name)) != NOTHING) {
    free_boolexp(DBFETCH(thing)->key);
    DBSTORE(thing, key, TRUE_BOOLEXP);
    notify(player, "Unlocked.");
  }
}

void do_unlink(dbref player, const char *name)
{
  dbref exit;
  struct match_data md;
  
  init_match(player, name, TYPE_EXIT, &md);
  match_all_exits(&md);
  match_here(&md);
  if(Wizard(player)) {
    match_absolute(&md);
  }
  
  switch(exit = match_result(&md)) {
  case NOTHING:
    notify(player, "Unlink what?");
    break;
  case AMBIGUOUS:
    notify(player, "I don't know which one you mean!");
    break;
  default:
    if(!controls(player, exit)) {
      notify(player, "Permission denied.");
    } else {
      switch(Typeof(exit)) {
      case TYPE_EXIT:
        if(DBFETCH(exit)->sp.exit.ndest != 0) {
	    DBFETCH(OWNER(exit))->sp.player.pennies += LINK_COST;
	    DBDIRTY(OWNER(exit));
	}
	DBSTORE(exit, sp.exit.ndest, 0);
	if (DBFETCH(exit)->sp.exit.dest) {
	  free ((void *)DBFETCH(exit)->sp.exit.dest);
	  DBSTORE(exit, sp.exit.dest, NULL);
	}
	notify(player, "Unlinked.");
	break;
      case TYPE_ROOM:
	DBSTORE(exit, sp.room.dropto, NOTHING);
	notify(player, "Dropto removed.");
	break;
      default:
	notify(player, "You can't unlink that!");
	break;
      }
    }
  }
}

#ifdef PLAYER_CHOWN
void do_chown(dbref player, const char *name, const char *newowner)
{
  dbref thing;
  dbref owner;
  struct match_data md;
  
  if (!*name) {
    notify(player, "You must specify what you want to take ownership of.");
    return;
  }
  
  init_match(player, name, NOTYPE, &md);
  match_everything(&md);
  if((thing = noisy_match_result(&md)) == NOTHING) return;
  
  if(*newowner && string_compare(newowner, "me")) {
    if ((owner = lookup_player(newowner)) == NOTHING) {
      notify(player, "I couldn't find that player.");
      return;
    }
  }
  else {
    owner = player;
  }
  
  if(!Wizard(player) && player != owner) {
    notify(player, "Only wizards can transfer ownership to others.");
    return;
  }
  
  if (!Wizard(player) && (!(FLAGS(thing) & CHOWN_OK)
			  || Typeof(thing) == TYPE_PROGRAM)) {
    notify(player, "You can't take possession of that.");
    return;
  }
  
  switch (Typeof(thing)) {
  case TYPE_ROOM:
    if (!Wizard(player) && DBFETCH(player)->location != thing) {
      notify(player, "You can only chown \"here\".");
      return;
    }
    OWNER(thing) = owner;
    break;
  case TYPE_THING:
    if (!Wizard(player) && DBFETCH(thing)->location != player) {
      notify(player, "You aren't carrying that.");
      return;
    }
    OWNER(thing) = owner;
    break;
  case TYPE_PLAYER:
    notify(player, "Players always own themselves.");
    return;
  case TYPE_EXIT:
  case TYPE_PROGRAM:
    OWNER(thing) = owner;
    break;
#ifdef RECYCLE
  case TYPE_GARBAGE:
    notify(player, "No one wants to own garbage.");
    return;
#endif	
  }
  if (owner == player)
    notify(player, "Owner changed to you.");
  else {
    char buf[BUFFER_LEN];
    sprintf(buf, "Owner changed to %s.", unparse_object(player, owner));
    notify(player, buf);
  }
  DBDIRTY(thing);
}

#else    /* without PLAYER_CHOWN */
void do_chown(dbref player, const char *name, const char *newowner)
{
  dbref thing;
  dbref owner;
  struct match_data md;
  
  if(!Wizard(player)) {
    notify(player, "Permission denied.");
    return;
  }
  
  init_match(player, name, NOTYPE, &md);
  match_everything(&md);
  if((thing = noisy_match_result(&md)) == NOTHING) return;
  if((owner = lookup_player(newowner)) == NOTHING) {
    notify(player, "I couldn't find that player.");
    return;
  }
  switch (Typeof(thing)) {
  case TYPE_PLAYER:
    notify(player, "Players always own themselves.");
    return;
  case TYPE_ROOM:
  case TYPE_THING:
  case TYPE_EXIT:
  case TYPE_PROGRAM:
    OWNER(thing) = owner;
#ifdef RECYCLE
  case TYPE_GARBAGE:
    notify(player, "No one wants to own garbage.");
    return;
#endif	
  }
  notify(player, "Owner changed.");
  DBDIRTY(thing);
}

#endif /* PLAYER_CHOWN */

/* Note: Gender code taken out.  All gender references are now to be handled
   by property lists...
   Setting of flags and property code done here.  Note that the PROP_DELIMITER
   identifies when you're setting a property.
   A @set <thing>= :
   will clear all properties.
   A @set <thing>= type:
   will remove that property.
   A @set <thing>= type: class
   will add that property or replace it.                                    */

void do_set(dbref player, const char *name, const char *flag)
{
  dbref thing;
  const char *p;
  object_flag_type f;
  
  /* find thing */
  if((thing = match_controlled(player, name)) == NOTHING) return;
  
  /* move p past NOT_TOKEN if present */
  for(p = flag; *p && (*p == NOT_TOKEN || isspace(*p)); p++);
  
  /* Now we check to see if it's a property reference */
  /* if this gets changed, please also modify boolexp.c */
  if (index(flag, PROP_DELIMITER))
    {
      /* copy the string so we can muck with it */
      char    *type = alloc_string(flag);   /* type */
      char    *class = (char *)index(type, PROP_DELIMITER);  /* class */
      char    *x; /* to preserve string location so we can free it */
      char    *temp;
      struct plist *prop, *last;
      
      x = type;
      while (isspace(*type) && (*type != PROP_DELIMITER))
	type++;
      if (*type == PROP_DELIMITER)
	{
	  /* clear all properties */
	  for (last = prop = DBFETCH(thing)->properties; prop; prop = last)
	    {
	      last = prop -> next;
	      free_prop(prop);
	    }
	  notify(player, "All properties removed.");
	  free((void *) x);
	  DBSTORE(thing, properties, NULL);
	  return;
	}
      else
	{
	  /* get rid of trailing spaces */
	  for (temp = class - 1; isspace(*temp); temp--)
	    ;
	  temp++;
	  *temp = '\0';
	}
      class++; /* move to next character */
      while (isspace(*class) && *class)
	class++;
      if (!(*class))
	{
	  remove_property(thing, type);
	  notify(player, "Property removed.");
	}
      else
	{
	  add_property(thing, type, class, 0);
	  notify(player, "Property set.");
	}
      free((void *) x);
      return;
    }
  
  /* identify flag */
  if(*p == '\0') {
    notify(player, "You must specify a flag to set.");
    return;
  } else if (string_prefix("WIZARD", p)) {
    f = WIZARD;
  } else if (string_prefix("LINK_OK", p)) {
    f = LINK_OK;
  } else if ((string_prefix("DARK", p)) || (string_prefix("DEBUG", p))) {
    f = DARK;
  } else if ((string_prefix("STICKY", p)) || (string_prefix("SETUID", p))) {
    f = STICKY;
#ifdef RESTRICTED_BUILDING
  } else if (string_prefix("BUILDER", p)) {
    f = BUILDER;
#endif	/* RESTRICTED_BUILDING */
#ifdef PLAYER_CHOWN
  } else if (string_prefix("CHOWN_OK", p)) {
    f = CHOWN_OK;
#endif /* PLAYER_CHOWN */
  } else if (string_prefix("JUMP_OK", p)) {
    f = JUMP_OK;
#ifdef HAVEN
  } else if (string_prefix("HAVEN", p)) {
    f = HAVEN;
#endif /* HAVEN */
#ifdef ABODE
  } else if (string_prefix("ABODE", p)) {
    f = ABODE;
#endif /* ABODE */
#ifndef MUCKER_ALL
  } else if (string_prefix("MUCKER", p)) {
    f = MUCKER;
#endif /* MUCKER_ALL */
  } else {
    notify(player, "I don't recognize that flag.");
    return;
  }
  
  /* check for restricted flag */
  if (restricted(player, thing, f)) {
    notify(player, "Permission denied.");
    return;
  }
  
  /* check for stupid wizard */
  if(f == WIZARD && *flag == NOT_TOKEN && thing == player) {
    notify(player, "You cannot make yourself mortal.");
    return;
  }
  
  /* else everything is ok, do the set */
  if(*flag == NOT_TOKEN) {
    /* reset the flag */
    FLAGS(thing) &= ~f;
    DBDIRTY(thing);
    notify(player, "Flag reset.");
  } else {
    /* set the flag */
    FLAGS(thing) |= f;
    DBDIRTY(thing);
    notify(player, "Flag set.");
  }
}

