/* $Header: /belch_a/users/rearl/tinymuck/src/RCS/wiz.c,v 1.16 90/09/28 12:25:37 rearl Exp $ */

/*
 * $Log:	wiz.c,v $
 * Revision 1.16  90/09/28  12:25:37  rearl
 * Fixed missing newline bug in @newpassword logging.
 * 
 * Revision 1.15  90/09/18  08:02:56  rearl
 * Fixed @tel for rooms -- a bug in permissions checking.
 * 
 * Revision 1.14  90/09/16  04:43:20  rearl
 * Preparation code added for disk-based MUCK.
 * 
 * Revision 1.13  90/09/15  22:28:34  rearl
 * Send inventory of the toad home, not the wizard's!
 * 
 * Revision 1.12  90/09/13  06:30:20  rearl
 * @toad modified to chown the victim's items to a recipient player.
 * 
 * Revision 1.11  90/09/10  02:19:06  rearl
 * Changed NL line termination to CR/LF pairs.
 * 
 * Revision 1.10  90/09/05  02:32:31  rearl
 * Added match_here() for room parent setting.
 * 
 * Revision 1.9  90/09/01  06:00:02  rearl
 * Fixed code in @teleport.
 * 
 * Revision 1.8  90/08/27  03:35:41  rearl
 * Changed teleport checks...
 * 
 * Revision 1.7  90/08/11  04:12:47  rearl
 * *** empty log message ***
 * 
 * Revision 1.6  90/08/06  03:49:14  rearl
 * Added logging of @force, @boot, and @toad.
 * 
 * Revision 1.5  90/08/05  03:20:16  rearl
 * Redid matching routines.
 * 
 * Revision 1.4  90/08/02  22:07:04  rearl
 * Changed one call to a log function, that's it.
 * 
 * Revision 1.3  90/07/29  17:46:28  rearl
 * Made @stat command a little cleaner, toaded victims are now 
 * toaded first, then all their connections are booted from the game.
 * 
 * Revision 1.2  90/07/23  14:48:37  casie
 * *** empty log message ***
 * 
 * Revision 1.1  90/07/19  23:04:20  casie
 * Initial revision
 * 
 *
 */

#include "copyright.h"
#include "config.h"

/* Wizard-only commands */

#include "db.h"
#include "params.h"
#include "interface.h"
#include "match.h"
#include "externs.h"

void do_teleport(dbref player, const char *arg1, const char *arg2)
{
  dbref victim;
  dbref destination;
  const char *to;
  struct match_data md;
  
  /* get victim, destination */
  if(*arg2 == '\0') {
    victim = player;
    to = arg1;
  } else {
    init_match(player, arg1, NOTYPE, &md);
    match_neighbor(&md);
    match_possession(&md);
    match_me(&md);
    match_here(&md);
    match_absolute(&md);
    match_player(&md);
    
    if((victim = noisy_match_result(&md)) == NOTHING) {
      return;
    }
    to = arg2;
  }
  
  /* get destination */
  init_match(player, to, TYPE_PLAYER, &md);
  match_here(&md);
  match_home(&md);
  match_absolute(&md);
  if(Wizard(player)) {
    match_neighbor(&md);
    match_me(&md);
    match_player(&md);
  }
  
  switch(destination = match_result(&md)) {
  case NOTHING:
    notify(player, "Send it where?");
    break;
  case AMBIGUOUS:
    notify(player, "I don't know which destination you mean!");
    break;
  case HOME:
    switch (Typeof(victim)) {
    case TYPE_PLAYER:
      destination = DBFETCH(victim)->sp.player.home;
      break;
    case TYPE_THING:
      destination = DBFETCH(victim)->sp.thing.home;
      break;
    case TYPE_ROOM:
      destination = GLOBAL_ENVIRONMENT;
      break;
    case TYPE_PROGRAM:
      destination = OWNER(victim);
      break;
    default:
      destination = PLAYER_START;  /* caught in the next switch anyway */
      break;
    }
  default:
    switch (Typeof(victim)) {
    case TYPE_PLAYER:
      if (!Wizard(player)) {
	notify(player, "Permission denied.");
	break;
      }
      if (Typeof(destination) != TYPE_ROOM) {
	notify(player, "Bad destination.");
	break;
      }
      notify(victim, "You feel a wrenching sensation...");
      enter_room(victim, destination, DBFETCH(victim)->location);
      notify(player, "Teleported.");
      break;
    case TYPE_THING:
    case TYPE_PROGRAM:
      if (Typeof(destination) != TYPE_ROOM
	  && Typeof(destination) != TYPE_PLAYER) {
	notify(player, "Bad destination.");
	break;
      }
      if (!Wizard(player)
	  && !((controls(player, destination)
		|| can_link_to(player, NOTYPE, destination))
	       && (controls(player, victim)
		   || controls(player, DBFETCH(victim)->location)))) {
	notify(player, "Permission denied.");
	break;
      }
      /* check for non-sticky dropto */
      if (Typeof(destination) == TYPE_ROOM
	  && DBFETCH(destination)->sp.room.dropto != NOTHING
	  && !(FLAGS(destination) & STICKY))
	destination = DBFETCH(destination)->sp.room.dropto;
      moveto(victim, destination);
      notify(player, "Teleported.");
      break;
    case TYPE_ROOM:
      if (Typeof(destination) != TYPE_ROOM) {
	notify(player, "Bad destination.");
	break;
      }
      if (!controls(player, victim)
	  || !can_link_to(player, NOTYPE, destination)
	  || victim == GLOBAL_ENVIRONMENT) {
	notify(player, "Permission denied.");
	break;
      }
      if (parent_loop_check(victim, destination)) {
	notify(player, "Parent would create a loop.");
	break;
      }
      moveto(victim, destination);
      notify(player, "Parent set.");
      break;
#ifdef RECYCLE
    case TYPE_GARBAGE:
      notify(player, "That object is in a place where magic cannot reach it.");
      break;
#endif
    default:
      notify(player, "You can't teleport that.");
      break;
    }
    break;
  }
  return;
}

void do_force(dbref player, const char *what, char *command)
{
  dbref victim;
  
  if(!Wizard(player)) {
    notify(player, "Only Wizards may use this command.");
    return;
  }
  
  /* get victim */
  if((victim = lookup_player(what)) == NOTHING) {
    notify(player, "That player does not exist.");
    return;
  }
  
#ifdef GOD_PRIV
  if (God(victim))
    {
      notify(player, "You cannot force god to do anything.");
      return;
    }
#endif /* GOD_PRIV */
  
  log_status("FORCED: %s(%d) by %s(%d): %s\n", NAME(victim),
	     victim, NAME(player), player, command);
  /* force victim to do command */
  process_command(victim, command);
}

void do_stats(dbref player, const char *name)
{
  int rooms;
  int exits;
  int things;
  int players;
  int programs;
#ifdef RECYCLE    
  int garbage = 0;
#endif    
  int total;
  dbref i;
  dbref owner;
  char buf[BUFFER_LEN];
  
  if(!Wizard(player)) {
    sprintf(buf, "The universe contains %d objects.", db_top);
    notify(player, buf);
  } else {
    total = rooms = exits = things = players = programs = 0;
    if (name != NULL && *name != '\0') {
      owner = lookup_player(name);
      if (owner == NOTHING) {
	notify(player, "I can't find that player.");
	return;
      }
      for (i = 0; i < db_top; i++) {
	switch(Typeof(i)) {
	case TYPE_ROOM:
	  if (OWNER(i) == owner)	 total++, rooms++;
	  break;
	  
	case TYPE_EXIT:
	  if (OWNER(i) == owner)    total++, exits++;
	  break;
	  
	case TYPE_THING:
	  if (OWNER(i) == owner)   total++, things++;
	  break;
	  
	case TYPE_PLAYER:
	  if (i == owner)                      total++, players++;
	  break;
	  
	case TYPE_PROGRAM:
	  if (OWNER(i) == owner) total++, programs++;
	  break;
	  
	}
      }
    }
    else {
      for(i = 0; i < db_top; i++) {
	switch(Typeof(i)) {
	case TYPE_ROOM:
	  total++;
	  rooms++;
	  break;
	case TYPE_EXIT:
	  total++;
	  exits++;
	  break;
	case TYPE_THING:
	  total++;
	  things++;
	  break;
	case TYPE_PLAYER:
	  total++;
	  players++;
	  break;
	case TYPE_PROGRAM:
	  total++;
	  programs++;
	  break;
#ifdef RECYCLE
	case TYPE_GARBAGE:
	  total++;
	  garbage++;
	  break;
#endif /* RECYCLE */
	}
      }
    }
#ifndef RECYCLE
    sprintf(buf,
	    "%d object%s = %d room%s, %d exit%s, %d thing%s, %d program%s, %d player%s.",
	    total, (total==1) ? "" : "s",
	    rooms, (rooms==1) ? "" : "s",
	    exits, (exits==1) ? "" : "s",
	    things, (things==1) ? "" : "s",
	    programs, (programs==1) ? "" : "s",
	    players, (players==1) ? "" : "s");
#else
    sprintf(buf,
	    "%d object%s  =   %d room%s, %d exit%s, %d thing%s, %d program%s, %d player%s, %d garbage.",
	    total, (total==1) ? "" : "s",
	    rooms, (rooms==1) ? "" : "s",
	    exits, (exits==1) ? "" : "s",
	    things, (things==1) ? "" : "s",
	    programs, (programs==1) ? "" : "s",
	    players, (players==1) ? "" : "s",
	    garbage);
#endif /* RECYCLE */
    notify(player, buf);
  }
}

void do_boot(dbref player, const char *name)
{
  dbref victim;
  char buf[BUFFER_LEN];
  
  if(!Wizard(player)) {
    notify(player, "Only a Wizard can boot someone off.");
    return;
  }
  
  if((victim = lookup_player(name)) == NOTHING) {
    notify(player, "That player does not exist.");
    return;
  }
  
  if(Typeof(victim) != TYPE_PLAYER) {
    notify(player, "You can only boot players!");
  }
#ifdef GOD_PRIV
  else if(God(victim)) {
    notify(player, "You can't boot God!");
  }
#endif /* GOD_PRIV */
  else {
    notify(victim, "You have been booted off the game.");
    if (boot_off(victim)) {
      log_status("BOOTED: %s(%d) by %s(%d)\n", NAME(victim),
		 victim, NAME(player), player);
      sprintf(buf, "You booted %s off!", NAME(victim));
      notify(player, buf);
    } else {
      sprintf(buf, "%s is not connected.", NAME(victim));
      notify(player, buf);
    }
  }
}

void do_toad(dbref player, const char *name, const char *recip)
{
  dbref victim;
  dbref recipient;
  dbref stuff;
  char buf[BUFFER_LEN];
  
  if(!Wizard(player)) {
    notify(player, "Only a Wizard can turn a person into a toad.");
    return;
  }
  
  if((victim = lookup_player(name)) == NOTHING) {
    notify(player, "That player does not exist.");
    return;
  }
  
  if (!*recip) {
    recipient = GOD;
  } else {
    if ((recipient = lookup_player(recip)) == NOTHING
	|| recipient == victim) {
      notify(player, "That recipient does not exist.");
      return;
    }
  }
  
  if(Typeof(victim) != TYPE_PLAYER) {
    notify(player, "You can only turn players into toads!");
  } else if(Wizard(victim)) {
    notify(player, "You can't turn a Wizard into a toad.");
  } else {
    /* we're ok */
    /* do it */
    send_contents(player, HOME);
    for (stuff = 0; stuff < db_top; stuff++) {
      if (OWNER(stuff) == victim) {
	switch (Typeof(stuff)) {
	case TYPE_ROOM:
	case TYPE_THING:
	case TYPE_EXIT:
	case TYPE_PROGRAM:
	  OWNER(stuff) = recipient;
	  DBDIRTY(stuff);
	  break;
	}
      }
    }
    if(DBFETCH(victim)->sp.player.password) {
      free((void *) DBFETCH(victim)->sp.player.password);
      DBFETCH(victim)->sp.player.password = 0;
    }
    FLAGS(victim) = TYPE_THING;
    OWNER(victim) = player; /* you get it */
    DBFETCH(victim)->sp.thing.value = 1;	/* don't let him keep his immense wealth */
    
    /* notify people */
    notify(victim, "You have been turned into a toad.");
    sprintf(buf, "You turned %s into a toad!", NAME(victim));
    notify(player, buf);
    log_status("TOADED: %s(%d) by %s(%d)\n", NAME(victim),
	       victim, NAME(player), player);
    
    delete_player(victim);
    /* reset name */
    sprintf(buf, "a slimy toad named %s", NAME(victim));
    free((void *) NAME(victim));
    NAME(victim) = alloc_string(buf);
    DBDIRTY(victim);
    while (boot_off(victim))
      ;
  }
}

void do_newpassword(dbref player, const char *name, const char *password)
{
  dbref victim;
  char buf[BUFFER_LEN];
  
  if(!Wizard(player)) {
    notify(player, "Your delusions of grandeur have been duly noted.");
    return;
  } else if((victim = lookup_player(name)) == NOTHING) {
    notify(player, "No such player.");
  } else if(*password != '\0' && !ok_password(password)) {
    /* Wiz can set null passwords, but not bad passwords */
    notify(player, "Bad password");
#ifdef GOD_PRIV
  } else if (God(victim))
    {
      notify(player, "You can't change God's password!");
      return;
#endif /* GOD_PRIV */
    } else {
      /* it's ok, do it */
      if(DBFETCH(victim)->sp.player.password)
	free((void *) DBFETCH(victim)->sp.player.password);
      DBSTORE(victim, sp.player.password, alloc_string(password));
      notify(player, "Password changed.");
      sprintf(buf, "Your password has been changed by %s.", NAME(player));
      notify(victim, buf);
      log_status("NEWPASS'ED: %s(%d) by %s(%d)\n", NAME(victim), (int) victim,
		 NAME(player), (int) player);
    }
}

#ifdef REGISTRATION
void
  do_pcreate(dbref player, const char *user, const char *password)
{
  dbref newguy;
  char buf[BUFSIZ];
  
#ifdef GOD_PRIV
  if (!God(player)) {
    notify(player, "Permission denied.");
    return;
  }
#else /* GOD_PRIV */
  if (!Wizard(player)) {
    notify(player, "Permission denied.");
    return;
  }
#endif
  
  newguy = create_player (user, password);
  if (newguy == NOTHING) {
    notify(player, "Create failed.");
  } else {
    log_status("PCREATED %s(%d) by %s(%d)\n",
	       NAME(newguy), (int) newguy, NAME(player), (int) player);
    sprintf(buf,"Player %s created as object #%d.",user,newguy);
    notify(player,buf);
  }
}
#endif /* REGISTRATION */
